# SPDX-FileCopyrightText: 2025-present U.N. Owen <void@some.where>
#
# SPDX-License-Identifier: MIT

import pydagoras.node as Node
Node = node.Node

import pydagoras.dag_dot as DAG
DAG = dag_dot.DAG

import pydagoras.dag_dot as calc 
calc = dag_dot.calc



#test_dag.py

import unittest

try:
    from pydagoras.node import Node

except ModuleNotFoundError:
    # for running tests, if pydagoras is not installed
    import sys, os

    cwd = os.getcwd()
    src_path = f'{cwd}/../src'
    src_pydagoras_path = f'{cwd}/../src/pydagoras'

    sys.path.insert(0, src_path)
    sys.path.insert(0, src_pydagoras_path)

    from pydagoras.node import Node


class TestNode(unittest.TestCase):
    def test_node_initialization(self):
        node = Node('A')
        self.assertEqual(node.node_id, 'A')

    def test_node_value_assignment(self):
        node = Node('B')
        node.set_value(10)
        self.assertEqual(node.get_value(), 10)

    def test_tooltip_assignment(self):
        node = Node('C')
        node.set_tooltip('This is a tooltip')
        self.assertEqual(node.tooltip, 'This is a tooltip')

    def test_node_str_representation(self):
        node = Node('D', nodetype='in')
        node.set_value(5)
        expected_str = 'Node(id:D, type:in, value:5, calc:None, usedby:None)'
        self.assertEqual(str(node), expected_str)

    def test_node_pp_json(self):
        node = Node('E', nodetype='in', tooltip='Output node')
        node.set_value(20)
        expected_json = {
            "node_id": "E",
            "value": 20,
            "calc": None,
            "usedby": None,
            "nodetype": "in",
            "display_name": "E",
            "tooltip": "Output node"
        }
        self.assertEqual(node.pp_json(), expected_json)

if __name__ == '__main__':
    unittest.main()



# dag.py

class DAG: # implementation
    __shared_state = {}

    def __init__(self):
        self.__dict__ = self.__shared_state

        if hasattr(self,'values'):
            return

        self.values = {}

    def __str__(self):
        return f'DAG(values={self.values})'

    def pp_json(self):
        return self.values
    
#    def __repr__(self):
#        return self.__str__()


    def set_value(self, key, value):
        self.values[key] = value
        
    def get_value(self, key):
        return self.values.get(key, None)

if __name__ == '__main__':
    print('dag.py main')
    my_dag = DAG()
    my_dag.set_value('a', 1)
    my_dag.set_value('b', 2)
     
    #my_dag.values['c'] = 3  # Directly setting a value
    #print(DAG().values)  # Output: {'a': 1, 'b': 2}

    #my_dag.test = 'test_value'  # Adding a new attribute dynamically
    #print(my_dag.test)  # Output: test_value
    #print(DAG().test)  # Output: test_value
    print('prints --------------')
    print (f'{my_dag=}')
    import json
    print(json.dumps(my_dag.pp_json(), sort_keys=True, indent=2))
    print(repr(my_dag))


# dag_dot.py

import logging
from pydagoras.dag import DAG

logger = logging.getLogger()


class Node:

    def __init__(self, node_id=None, calc=None, usedby=None, nodetype=None, display_name=None, tooltip='notset'):
        self.calc = calc
        self.node_id = node_id
        self.usedby = usedby
        self.dag = DAG()   
        self.dag.set_value(node_id, None)
        self.nodetype = nodetype
    
        if display_name:
            self.display_name = display_name
        else:
            self.display_name = node_id

        self.tooltip = tooltip # use tooltip to show errors when present
        self.orig_tooltip = tooltip


    def __str__(self):
        return f'Node(id:{self.node_id}, type:{self.nodetype}, value:{self.get_value()}, calc:{self.calc}, usedby:{[n.node_id for n in self.usedby] if self.usedby else None})'

    def pp_json(self): 
        return { "node_id": self.node_id, \
                 "value": self.get_value(), \
                 "calc": self.calc, \
                 "usedby": [n.node_id for n in self.usedby] if self.usedby else None, \
                 "nodetype": self.nodetype, \
                 "display_name": self.display_name, \
                 "tooltip": self.tooltip}
        
    def __repr__(self):
        return f'Node({str(self.pp_json())})'

    def pp(self):
        print(f'NODE: {self.nodetype}, id:{self.node_id}, value:{self.get_value()}')
        print(f'      display_name:{self.display_name} tooltip:{self.tooltip} ')
        print(f'      calc: {self.calc} usedby:{self.usedby[0].node_id if self.usedby else None}')


    def set_tooltip(self, tooltip):
        self.tooltip = tooltip


    def set_value(self, value):
        self.dag.set_value(self.node_id, value)


    def get_value(self):
        return self.dag.get_value(self.node_id)


if __name__ == '__main__':
    print('##########################################')
    def calc_simple(node=None):
        # calc_simple
        return DAG().get_value(node.node_id) + 2

    my_node = Node(node_id='a', nodetype='in')
    my_node.dag.set_value('a', 1)
    my_node.pp()  # Output: Input a = 1
    
    import json
    print( json.dumps(my_node.pp_json(), indent=2) )
    print( repr(my_node))



# filepath: /Users/python/pydagoras/dev/2026-09-06_tests/pydagoras/src/pydagoras/test_dag.py

import os
import sys
import json
import unittest

#jfrom pydagoras.dag_dot import DAG_dot, calc
#from pydagoras.dag import DAG

try:
    from pydagoras.dag import DAG
    from pydagoras.dag_dot import DAG_dot, calc

    #from pydagoras.node import Node
    #from pydagoras.dag_dot import DAG_dot, calc

except ModuleNotFoundError:
    # for running tests, if pydagoras is not installed
    import sys, os

    cwd = os.getcwd()
    src_path = f'{cwd}/../src'
    src_pydagoras_path = f'{cwd}/../src/pydagoras'

    sys.path.insert(0, src_path)
    sys.path.insert(0, src_pydagoras_path)

    from pydagoras.dag import DAG
    from pydagoras.dag_dot import DAG_dot, calc


class TestDAG(unittest.TestCase):
    def setUp(self):
        # Ensure a fresh shared state for each test
        DAG._DAG__shared_state.clear()

    def test_set_and_get_value(self):
        dag = DAG()
        dag.set_value('x', 42)
        self.assertEqual(dag.get_value('x'), 42)

    def test_get_value_missing_key(self):
        dag = DAG()
        self.assertIsNone(dag.get_value('missing'))

    def test_shared_state_across_instances(self):
        dag1 = DAG()
        dag2 = DAG()
        dag1.set_value('shared', 'yes')
        self.assertEqual(dag2.get_value('shared'), 'yes')

    def test_direct_values_assignment(self):
        dag = DAG()
        dag.values['direct'] = 'set'
        self.assertEqual(dag.get_value('direct'), 'set')

    def test_dynamic_attribute_shared(self):
        dag1 = DAG()
        dag1.custom_attr = 'custom'
        dag2 = DAG()
        self.assertTrue(hasattr(dag2, 'custom_attr'))
        self.assertEqual(dag2.custom_attr, 'custom')

    def test_values_dict_is_shared(self):
        dag1 = DAG()
        dag2 = DAG()
        dag1.set_value('a', 1)
        dag2.set_value('b', 2)
        self.assertEqual(dag1.values, dag2.values)
        self.assertEqual(dag1.values, {'a': 1, 'b': 2})

    def test_str(self):
        dag = DAG()
        dag.set_value('a', 1)
        dag.set_value('b', 2)
        self.assertEqual(str(dag), "DAG(values={'a': 1, 'b': 2})")
       
        dag_json = json.dumps(dag.values, sort_keys=True, indent=2)
        print(dag_json)
        print(str(dag_json))

    def test_pp_json(self):
        dag = DAG()
        dag.set_value('a', 1)
        dag.set_value('b', 2)
        expected_json = {
            "a": 1,
            "b": 2
        }
        self.assertEqual(dag.pp_json(), expected_json)  

if __name__ == '__main__':
    unittest.main()



# dag_dot.py

from functools import wraps

import logging
import pygraphviz as pgv
from pydagoras.dag import DAG
from pydagoras.node import Node

logger = logging.getLogger(__name__)


class DAG_dot:

    def __init__(self, label):
        logger.info(f'creating {label=}')
        self.label = label
        self.dag = DAG()  # Initialize the DAG instance
        self.G=pgv.AGraph(directed=True, strict=True, rankdir='LR', label=label, labelloc="t")
        self.input_nodes = []
        self.internal_nodes = []
        self.output_node = None
        self.nodes = []  # List to keep track of all nodes


        @calc
        def calc_out(node=None):
            return self.n_out.get_value() 

        self.n_out = self.makeNode('out', calc=calc_out, usedby=[], nodetype='out') 
        
    def __str__(self):
        return f'DAG_dot(label:{self.label}, nodes:{len(self.nodes)})'


    def makeNode(self,label,calc,usedby=None, nodetype='internal', display_name=None, tooltip=''):
        if usedby is None:
            usedby = [self.n_out]
        n = Node(label,calc,usedby,nodetype,display_name,tooltip)
        if nodetype == 'in':
            self.input_nodes.append(n)
        elif nodetype == 'out':
            self.output_node = n
        else:
            self.internal_nodes.append(n)
        self.nodes.append(n)  # Add the node to the list of all nodes

        self.drawNode(n,usedby, nodetype, tooltip)
        return n


    def drawNode(self,node,usedby,nodetype,tooltip):
        doc = node.display_name
        if nodetype == 'in':
            self.G.add_node(doc, shape="square", tooltip=tooltip)
            for n in usedby:
                self.draw_edge(doc,n.display_name)
        elif nodetype == 'internal':
            self.G.add_node(doc, tooltip=tooltip)
            for n in usedby:
                self.draw_edge(doc,n.display_name)
        elif nodetype == 'out':
            self.G.add_node(doc, color="white")


    def draw_edge(self,node1,node2):
        self.G.add_edge(node1,node2,label='Undefined', fontname="Courier")


    def update_node(self,node1,node2,value,tooltip='not set XXX'):
        fontcolor, color = self.get_colors(value)
        self.G.add_node(node1,color=color,fontcolor=fontcolor,tooltip=tooltip)
        self.G.add_edge(node1,node2, label=value,fontcolor=fontcolor,color=color, fontname="Courier")
 
    # special cases
    @classmethod
    def get_colors(cls, value):
        if value in ( 'e',):
            return 'red', 'red'
        return 'blue', 'green'

    def set_input(self,node_id,value): # set values in all node
        for node in self.nodes:
            if node.node_id == node_id:
                for usedby in node.usedby:
                    self.update_node(node.display_name,usedby.node_id, value=value, tooltip=node.tooltip)
                    self.set_input(usedby.node_id, value)  # recursion 
                    #self.setValue(node, value)
                self.setValue(node, value)


    def setValue(self,n,v):
        if v == self.dag.get_value(n):
            return

        # build the DAG
        n.set_value(v) 
        x = self.dag.get_value(n.node_id)

        for u in n.usedby:
            if u.calc == None:
                continue
            
            new_value = u.calc(self, node=n)  # Call the calc function with self and node, contains error handling
            self.dag.set_value(u.node_id, new_value)

        if not n.usedby:
            return

        if n.usedby[0].usedby == []: # output node
            msg = 'update dag_dot.py %s %s' %(n.usedby[0].node_id, n.get_value())
            logger.info (msg)


    def ppInputs(self):
        for n in self.input_nodes:
            print(n)

    def ppInternals(self):
        for n in self.internal_nodes:
            print(n)

    def ppOutput(self):
        print(self.output_node)

    def pp(self):
        print(f'DAG: {self.label}')
        print('Inputs:')
        self.ppInputs()
        print('Internals:')
        self.ppInternals()
        print('Output:')
        self.ppOutput()

    def pp_json(self):
        values = self.dag.values  # Print the values of all nodes in the DAG
        return {
            "label": self.label,
            "values": values
        }



    def get_value(self):
         return self.dag.get_value(self.label)

    def set_value(self, node, value):
        self.dag.set_value(node.node_id, value)
        self.setValue(node,value)


def calc(f1): # decorator deffinition
    @wraps(f1)
    def f3(dag, *args, **kwargs):
        node=kwargs['node']

        u_node = node.usedby[0] if node.usedby else None
        #        self.update_node(u_node.node_id,o_node.node_id, value='-', tooltip=node.tooltip)

        try:
            rtn = f1(*args, **kwargs) # Call the original function
            u_node.set_value(rtn)  # Set the value of the node
            u_node.set_tooltip(u_node.orig_tooltip)  # Set the tooltip of the node

        except Exception as e:
            logger.error('Error in %s: %s' %(u_node.node_id,str(e)))
            rtn = 'e'
            u_node.set_value(rtn)  # Set the value of the node
            u_node.set_tooltip(str(e))  # Set the tooltip of the node

        for u_node in node.usedby:
            #print(f'Updating node: {u_node.node_id} with value: {rtn}')
            dag.set_input(u_node.node_id, rtn)

        return rtn
    return f3

if __name__ == '__main__':

    @calc
    def calc_triple(node=None):
        return node.get_value() * 3 

    my_dag = DAG_dot(label='Test DAG')
    n2 = my_dag.makeNode('b', calc=calc_triple)
    n1 = my_dag.makeNode('a', calc=None, usedby=[n2], nodetype='in') #, display_name='Input A')
    my_dag.set_input('a', 10)
    
    print(f'{my_dag=}')  

    my_dag.pp()

    import json
    print(json.dumps(my_dag.pp_json(), sort_keys=True, indent=2))
    
    print('Graph representation:')  
    print(my_dag.G.to_string())  # Print the graph representation  

    print(repr(my_dag))  
    


# tests/test_dag_dot.py

import unittest

try:
    from pydagoras.node import Node
    from pydagoras.dag_dot import DAG_dot, calc

except ModuleNotFoundError:
    # for running tests, if pydagoras is not installed
    import sys, os

    cwd = os.getcwd()
    src_path = f'{cwd}/../src'
    src_pydagoras_path = f'{cwd}/../src/pydagoras'

    sys.path.insert(0, src_path)
    sys.path.insert(0, src_pydagoras_path)

    from dag_dot import DAG_dot, calc


class TestDAGDot(unittest.TestCase):
    def setUp(self):
        self.dag_dot = DAG_dot(label='Test DAG')


    def test_makeNode_adds_node(self):
        n = self.dag_dot.makeNode('test', calc=None)
        self.assertIn(n, self.dag_dot.nodes)
        self.assertTrue(any(node.node_id == 'test' for node in self.dag_dot.nodes))

    def test_input_and_output_nodes(self):
        n_in = self.dag_dot.makeNode('input', calc=None, nodetype='in')
        n_out = self.dag_dot.makeNode('output', calc=None, nodetype='out')
        self.assertIn(n_in, self.dag_dot.input_nodes)
        self.assertEqual(self.dag_dot.output_node, n_out)

    def test_drawNode_adds_to_graph(self):
        n = self.dag_dot.makeNode('draw', calc=None)
        self.assertIn(n.display_name, self.dag_dot.G.nodes())

    def test_set_and_get_value(self):
        n = self.dag_dot.makeNode('val', calc=None)
        self.dag_dot.set_value(n, 123)
        self.assertEqual(self.dag_dot.dag.get_value(n.node_id), 123)

    def test_get_colors(self):
        self.assertEqual(DAG_dot.get_colors('e'), ('red', 'red'))
        self.assertEqual(DAG_dot.get_colors('anything'), ('blue', 'green'))

    def test_calc_decorator_success(self):

        @calc
        def triple(node=None):
            return node.get_value() * 3

        my_dag = DAG_dot(label='Eg DAG')
        n2 = my_dag.makeNode('x3', calc=triple, tooltip='multiply')
        my_dag.makeNode('In', calc=None, usedby=[n2], nodetype='in')

        my_dag.set_input('In', 10)

        result = my_dag.output_node.get_value()
        self.assertEqual(result, 30)

    def test_calc_decorator_exception(self):

        @calc
        def fail(node=None):
            raise ValueError("fail")

        my_dag = DAG_dot(label='Eg DAG')
        n2 = my_dag.makeNode('Eg fail', calc=fail, tooltip='eg_error')
        n1 = my_dag.makeNode('In', calc=None, usedby=[n2], nodetype='in')
        my_dag.set_input('In', 10)

        self.assertEqual(n2.tooltip, 'fail')

        #print(my_dag.G.to_string())  # Print the graph representation  


if __name__ == '__main__':
    unittest.main()



# eg_use_pydagoras.py
# a script to provide an example of creating and using a DAG using pydagoras


import os
import sys

try:
    from pydagoras.dag_dot import DAG_dot, calc
except ModuleNotFoundError as e:
    cwd = os.getcwd()
    src_path = f'{cwd}/../src'
    src_pydagoras_path = f'{cwd}/../src/pydagoras'

    sys.path.insert(0, src_path)
    sys.path.insert(0, src_pydagoras_path)

    from dag import DAG
    from dag_dot import DAG_dot, calc

def run():

    print('#######################################')

    @calc
    def tripple(node=None):
        return node.get_value() * 3

    dag = DAG_dot(label='Eg DAG')
    n2 = dag.makeNode('x3', calc=tripple, tooltip='multiply')
    n1 = dag.makeNode('In', calc=None, usedby=[n2], nodetype='in')

    print('Initial DAG')
    print(dag.G.to_string()) # 
    print('Updates --------------')
    dag.set_input('In', 10)

    print('Outputs --------------')
    dag.ppInputs() # 
    dag.ppOutput() # 
    dag.pp()

    print('Final DAG')
    print(dag.G.to_string()) # 

if __name__ == '__main__':
    run()
    print('Done.')
